-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local _, L = ...;

local bitsSet = {1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4};

local DEFAULT_TEXTURE_COLOUR = {r = 1.0, g = 1.0, b = 1.0, a = 1.0};

local NIGGLES_INVITESTATUS_UNKNOWN = (CALENDAR_MAX_INVITESTATUS+1);

local onlineStatusInfo =
{
  [0] =
  {
    texture = READY_CHECK_NOT_READY_TEXTURE,
    colour  = DEFAULT_TEXTURE_COLOUR,
    coord   = {0.0, 1.0, 0.0, 1.0}
  },
  [1] =
  {
    texture = READY_CHECK_READY_TEXTURE,
    colour  = DEFAULT_TEXTURE_COLOUR,
    coord   = {0.0, 1.0, 0.0, 1.0}
  },
  [2] =
  {
    texture = "Interface\\CHATFRAME\\UI-ChatIcon-ArmoryChat",
    colour  = {r = 0.0, g = 1.0, b = 0.0},
    coord   = {0.0, 1.0, 0.0, 1.0}
  },
  [3] =
  {
    texture = "Interface\\CHATFRAME\\UI-ChatIcon-ArmoryChat-AwayMobile",
    colour  = DEFAULT_TEXTURE_COLOUR,
    coord   = {0.0, 1.0, 0.0, 1.0}
  },
  [4] =
  {
    texture = "Interface\\CHATFRAME\\UI-ChatIcon-ArmoryChat-BusyMobile",
    colour  = DEFAULT_TEXTURE_COLOUR,
    coord   = {0.0, 1.0, 0.0, 1.0}
  },
  [5] =
  {
    texture = READY_CHECK_WAITING_TEXTURE,
    colour  = DEFAULT_TEXTURE_COLOUR,
    coord   = {0.0, 1.0, 0.0, 1.0}
  },
  [6] =
  {
    texture = "Interface\\HELPFRAME\\HelpIcon-ItemRestoration",
    colour  = DEFAULT_TEXTURE_COLOUR,
    coord   = {0.15625, 0.875, 0.15625, 0.875}
  }
};

local roleInfo =
{
  [0] =
  {
    flag = 0,
    text = NONE,
    icon = {0.6875, 1.0,      0.0,      0.3125  }
  },
  [1] =
  {
    flag = 1,
    text = INLINE_TANK_ICON..TANK,
    icon = {0.0,    0.296875, 0.328125, 0.640625}
  },
  [2] =
  {
    flag = 2,
    text = INLINE_HEALER_ICON..HEALER,
    icon = {0.3125, 0.609375, 0.0,      0.3125  }
  },
  [3] =
  {
    flag = 4,
    text = INLINE_DAMAGER_ICON..DAMAGER,
    icon = {0.3125, 0.609375, 0.328125, 0.640625}
  },
};

local rolePref =
{
  {
    label = L["RolePrimary"],
    shift = 0,
    mask  = 0xFFF0,
  },
  {
    label = L["RoleSecondary"],
    shift = 4,
    mask  = 0xFF0F,
  },
};

-- Paths for the textures to display for calendar events
local calendarEventTypeTexture =
{
  [CALENDAR_EVENTTYPE_RAID] =
  {
    default  = "Interface\\LFGFrame\\LFGIcon-Raid",
    detailed = "Interface\\LFGFrame\\LFGIcon-"
  },
  [CALENDAR_EVENTTYPE_DUNGEON] =
  {
    default  = "Interface\\LFGFrame\\LFGIcon-Dungeon",
    detailed = "Interface\\LFGFrame\\LFGIcon-"
  },
  [CALENDAR_EVENTTYPE_PVP] =
  {
    default  = "Interface\\Calendar\\UI-Calendar-Event-PVP",
    detailed = ""
  },
  [CALENDAR_EVENTTYPE_MEETING] =
  {
    default = "Interface\\Calendar\\MeetingIcon",
    detailed = ""
  },
  [CALENDAR_EVENTTYPE_OTHER] =
  {
    default = "Interface\\Calendar\\UI-Calendar-Event-Other",
    detailed = ""
  }
};

-- Colours for calendar event titles
local calendarEventColor =
{
  [""]          = NORMAL_FONT_COLOR;
  ["CREATOR"]   = {r=0.54, g=0.75, b=1.0};
  ["MODERATOR"] = {r=0.54, g=0.75, b=1.0};
}

-- Text & colours for calendar event invite status
local inviteStatusInfo =
{
  [CALENDAR_INVITESTATUS_INVITED] =
  {
    text      = CALENDAR_STATUS_INVITED,
    color     = NORMAL_FONT_COLOR,
    colorCode = NORMAL_FONT_COLOR_CODE
  },
  [CALENDAR_INVITESTATUS_ACCEPTED] =
  {
    text      = CALENDAR_STATUS_ACCEPTED,
    color     = GREEN_FONT_COLOR,
    colorCode = GREEN_FONT_COLOR_CODE
  },
  [CALENDAR_INVITESTATUS_DECLINED] =
  {
    text      = CALENDAR_STATUS_DECLINED,
    color     = RED_FONT_COLOR,
    colorCode = RED_FONT_COLOR_CODE
  },
  [CALENDAR_INVITESTATUS_CONFIRMED] =
  {
    text      = CALENDAR_STATUS_CONFIRMED,
    color     = GREEN_FONT_COLOR,
    colorCode = GREEN_FONT_COLOR_CODE
  },
  [CALENDAR_INVITESTATUS_OUT] =
  {
    text      = CALENDAR_STATUS_OUT,
    color     = RED_FONT_COLOR,
    colorCode = RED_FONT_COLOR_CODE
  },
  [CALENDAR_INVITESTATUS_STANDBY] =
  {
    text      = CALENDAR_STATUS_STANDBY,
    color     = ORANGE_FONT_COLOR,
    colorCode = ORANGE_FONT_COLOR_CODE
  },
  [CALENDAR_INVITESTATUS_SIGNEDUP] =
  {
    text      = CALENDAR_STATUS_SIGNEDUP,
    color     = GREEN_FONT_COLOR,
    colorCode = GREEN_FONT_COLOR_CODE
  },
  [CALENDAR_INVITESTATUS_NOT_SIGNEDUP] =
  {
    text      = CALENDAR_STATUS_NOT_SIGNEDUP,
    color     = NORMAL_FONT_COLOR,
    colorCode = NORMAL_FONT_COLOR_CODE
  },
  [CALENDAR_INVITESTATUS_TENTATIVE] =
  {
    text      = CALENDAR_STATUS_TENTATIVE,
    color     = ORANGE_FONT_COLOR,
    colorCode = ORANGE_FONT_COLOR_CODE
  },
  [NIGGLES_INVITESTATUS_UNKNOWN] =
  {
    text      = UNKNOWN,
    color     = NORMAL_FONT_COLOR,
    colorCode = NORMAL_FONT_COLOR_CODE
  }
};

-- Order of invite status in the drop down menus
local inviteStatusOrder =
{
  CALENDAR_INVITESTATUS_NOT_SIGNEDUP,
  CALENDAR_INVITESTATUS_INVITED,
  CALENDAR_INVITESTATUS_SIGNEDUP,
  CALENDAR_INVITESTATUS_ACCEPTED,
  CALENDAR_INVITESTATUS_CONFIRMED,
  CALENDAR_INVITESTATUS_STANDBY,
  CALENDAR_INVITESTATUS_TENTATIVE,
  CALENDAR_INVITESTATUS_OUT,
  CALENDAR_INVITESTATUS_DECLINED
};

-- Children to take ownership of
local ownershipChildren =
{
  {name = "CalendarCreateEventFrame",   x = 75, y = 50},
  {name = "CalendarViewEventFrame",     x = 75, y = 50},
  {name = "CalendarTexturePickerFrame", x = 75, y = 50},
  {name = "CalendarContextMenu"                       }
};

-- Info for controlling ownership of the calendar event edit/view frames
local ownershipInfo = {};

-- List of all calendar player events, guild events and guild announcements
local calendarEvent = {};

-- List of the guild members used to indicate who is online
local guildRoster = {[0] = 0};

-- Frames to add role and online status info to invitees
-- NOTE: Only one set is used to save resources. They are reparented when
--       'CalendarCreateEventFrame' or 'CalendarViewEventFrame' are shown,
local roleStatusFrames = {};

local eventPanelInfo =
{
  [1] =
  {
    prefix         = "CalendarCreateEvent",
    panel          = nil,
    divider        = nil,
    list           = nil,
    scrollFrame    = nil,
    scrollPoints   = {{},{}},
    updateName     = "NigglesCreateEventInviteListUpdate",
    updateOrigName = "CalendarCreateEventInviteListScrollFrame_Update",
    updateOrigFunc = nil,
    yOffset        = 23,
    filterBtn      = {},
    filteredBar    = nil,
    headingClass   = nil,
    headingRoles   = nil,
    labelInvites   = nil,
  },
  [2] =
  {
    prefix         = "CalendarViewEvent",
    panel          = nil,
    divider        = nil,
    list           = nil,
    scrollFrame    = nil,
    scrollPoints   = {{},{}},
    updateName     = "NigglesViewEventInviteListUpdate",
    updateOrigName = "CalendarViewEventInviteListScrollFrame_Update",
    updateOrigFunc = nil,
    yOffset        = 4,
    filterBtn      = {},
    filteredBar    = nil,
    headingClass   = nil,
    headingRoles   = nil,
    labelInvites   = nil,
  }
};

local sortedCharRoles  = {};
local filteredInvites  = {[0] = 0};
local filterRoles      = bit.lshift(1, #roleInfo)-1;
local currentEventTime = 0;

-------------------------------------------------------------------------------
--              G  L  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

StaticPopupDialogs["NIGGLES_CALEVENT_REMOVE"] =
{
  text         = L["CalEventRemove"],
  button1      = YES,
  button2      = NO,
  whileDead    = 1,
  OnAccept     = function(self, data)
    CalendarEventRemoveInvite(data);
  end,
  showAlert    = 1,
  timeout      = 0,
  exclusive    = 1,
  hideOnEscape = 1
};

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local _L = {};
local calendarEventGetTexture;
local calendarEventsUpdate;
local characterRolesGetSortedList;
local characterRolesScrollFrameUpdate;
local filteredInvitesUpdate;
local inviteFileredBarOnClick;
local inviteFilterOnClick;
local inviteFilterOnEvent;
local roleGetTexCoord;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to get the list of calendar events
--
local function calendarEventsGetList()
  -- Local Variables
  local currentEvent;
  local eventCalType;
  local eventDate = {};
  local eventInviteStatus;
  local eventModStatus;
  local eventTexture;
  local eventTitle;
  local eventType;
  local monthOffset = 0;
  local numDays;
  local numEvents;

  -- Clear any existing calendar events
  wipe(calendarEvent);

  -- Get the current date
  _, eventDate.month, eventDate.day, eventDate.year = CalendarGetDate();

  -- Update the day displayed in the portrait
  NigglesCalendarEvents.day:SetText(eventDate.day);

  -- Get the number of days in the current month
  _, _, numDays = CalendarGetMonth();

  -- Get all the events from today for the next 28 days
  for dayIdx = 1, 28 do
    numEvents = CalendarGetNumDayEvents(monthOffset, eventDate.day);
    for eventIdx = 1, numEvents do
      eventTitle, eventDate.hour, eventDate.min, eventCalType, _, eventType,
      eventTexture, eventModStatus, eventInviteStatus =
        CalendarGetDayEvent(monthOffset, eventDate.day, eventIdx);
      if ((eventCalType == "GUILD_ANNOUNCEMENT") or
          (eventCalType == "GUILD_EVENT") or
          (eventCalType == "PLAYER")) then
        calendarEvent[#calendarEvent+1] = {};
        currentEvent = calendarEvent[#calendarEvent];
        currentEvent.index       = eventIdx;
        currentEvent.day         = eventDate.day;
        currentEvent.monthOffset = monthOffset;
        currentEvent.date        = date("%A, %Y-%m-%d, %H:%M", time(eventDate));
        currentEvent.title       = eventTitle;
        currentEvent.type        = eventType;
        currentEvent.texture     = eventTexture;
        currentEvent.modStatus   = eventModStatus;
        if (eventCalType ~= "GUILD_ANNOUNCEMENT") then
          currentEvent.inviteStatus = eventInviteStatus;
        else
          currentEvent.inviteStatus = nil;
        end
      end
    end

    -- Move on to the next day
    eventDate.day = eventDate.day+1;
    if (eventDate.day > numDays) then
      monthOffset   = monthOffset+1;
      eventDate.day = 1;
      eventDate.month, eventDate.year, numDays = CalendarGetMonth(monthOffset);
    end
  end

  -- Update the Calendar Events scroll frame
  calendarEventsUpdate();

  return;
end

--
-- Function to take ownership of the frames used to edit/view events
--
function calendarEventsTakeOwnership(frame)
  -- Give ownership of the Calendar's children to the frame
  if (ownershipInfo[frame] ~= nil) then
    for _, childInfo in ipairs(ownershipInfo[frame]) do
      childInfo.frame:SetParent(frame);
      childInfo.frame:ClearAllPoints();
      for _, point in ipairs(childInfo.points) do
        childInfo.frame:SetPoint(point[1], point[2], point[3],
          point[4], point[5]);
      end
    end
  end

  return;
end

--
-- Function to update the scroll frame in the Calendar Events panel
--
calendarEventsUpdate = function()
  -- Local Variables
  local button;
  local displayedHeight;
  local eventIdx;
  local totalHeight;
  local textColor;
  local currentEvent;
  local scrollFrame = NigglesCalendarEventsScrollFrame;
  local topOffset = HybridScrollFrame_GetOffset(scrollFrame);
  local numButtons = #scrollFrame.buttons;
  local numEvents = #calendarEvent;
  local statusInfo = nil;

  -- Make sure the top offset is valid
  if (topOffset > numEvents) then
    topOffset = numEvents;
    HybridScrollFrame_SetOffset(scrollFrame, topOffset);
  end

  -- Update the scroll frame's buttons
  for idx = 1, numButtons do
    button = scrollFrame.buttons[idx];
    eventIdx = topOffset+idx;
    if (eventIdx <= numEvents) then
      -- Assign the current event to a more convenient variable
      currentEvent = calendarEvent[eventIdx];

      -- Set the event's title
      button.title:SetText(currentEvent.title);
      textColor = calendarEventColor[currentEvent.modStatus];
      button.title:SetTextColor(textColor.r, textColor.g,
        textColor.b, textColor.a);

      -- Set the event's date/time
      button.date:SetText(currentEvent.date);

      -- Get the info for the event's invite status
      if ((currentEvent.inviteStatus ~= nil) and
          (currentEvent.inviteStatus <= CALENDAR_MAX_INVITESTATUS) and
          (inviteStatusInfo[currentEvent.inviteStatus] ~= nil)) then
        statusInfo = inviteStatusInfo[currentEvent.inviteStatus];
      else
        statusInfo = nil;
      end

      -- Set the event's invite status
      if (statusInfo ~= nil) then
        button.inviteStatus:SetText(statusInfo.colorCode..statusInfo.text);
        button.inviteStatus:Show();
      else
        button.inviteStatus:Hide();
      end

      -- Set the event's icon
      button.icon:SetTexture(calendarEventGetTexture(
        calendarEvent[eventIdx].type,
        calendarEvent[eventIdx].texture));

      -- Show/Hide the event's signup icon
      button.signup:SetShown(calendarEvent[eventIdx].inviteStatus == 1);

      -- Save the event the button is for
      button.event = currentEvent;

      -- Show the button
      button:Show();
    else
      -- Hide the button
      button:Hide();
    end
  end

  -- Update the scroll frame's height
  totalHeight     = numEvents*scrollFrame.buttons[1]:GetHeight();
  displayedHeight = numButtons*scrollFrame.buttons[1]:GetHeight();
  HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);

  return;
end

--
-- Function to work out what texture to display for an calendar event
--
calendarEventGetTexture = function(eventType, texture)
  -- Local Variables
  local texturePath =
    calendarEventTypeTexture[CALENDAR_EVENTTYPE_OTHER].default;

  -- Check if the event type has textures
  if (calendarEventTypeTexture[eventType] ~= nil) then
    -- Check if the event type has a detailed texture path
    if ((calendarEventTypeTexture[eventType].detailed ~= "") and
        (texture ~= nil) and
        (texture ~= "")) then
      texturePath = calendarEventTypeTexture[eventType].detailed..texture;
    elseif (calendarEventTypeTexture[eventType].default ~= "") then
      texturePath = calendarEventTypeTexture[eventType].default;
    end
  end

  return texturePath;
end

--
-- Function called when a Calendar Event panel is shown
--
local function calendarInvitesShow(panel)
  -- Local Variables
  local button;
  local buttons;
  local currentInfo;
  local displayInvites = NigglesDBC.settings.CalEventsInvites;
  local displayOnline = NigglesDBC.settings.CalEventsOnline;
  local displayRoles = NigglesDBC.settings.CalEventsRoles;
  local roleStatus;
  local numButtons;
  local child;
  local texture;

  -- Reset the filter flags
  filterRoles = bit.lshift(1, #roleInfo)-1;

  -- Get the info for the panel
  currentInfo = eventPanelInfo[(panel ~= nil) and panel:GetID() or 0];
  if (currentInfo == nil) then
    return;
  end

  -- Adjust the layout of the invite list
  if (currentInfo.list ~= nil) then
    currentInfo.list:SetPoint("TOPLEFT", currentInfo.divider,
      "BOTTOMLEFT", 3, -14-(displayRoles and 24 or 0));
  end

  -- Adjust the position of the scroll frame and scroll bar
  if (currentInfo.scrollFrame ~= nil) then
    if (displayRoles) then
      currentInfo.scrollFrame:SetPoint("TOPLEFT", 3, -19)
      currentInfo.scrollFrame.scrollBar:SetPoint("TOPLEFT",
        currentInfo.list, "TOPRIGHT", -25, -19)
    else
      NigglesFramePointsSet(currentInfo.scrollFrame,
        currentInfo.scrollPoints[1]);
      NigglesFramePointsSet(currentInfo.scrollFrame.scrollBar,
        currentInfo.scrollPoints[2]);
    end
  end

  -- Show/Hide the number of invites
  if ((displayInvites) or (currentInfo.labelInvites ~= 0)) then
    -- Create the 'Roles' heading, if required
    if (currentInfo.labelInvites == nil) then
      currentInfo.labelInvites = currentInfo.list:CreateFontString(nil,
        "ARTWORK", "GameFontNormalSmall");
      currentInfo.labelInvites:SetPoint("BOTTOMLEFT",
        _G[currentInfo.prefix.."InviteListNameSortButton"],
        "BOTTOMRIGHT", 0, 2);
    end

    -- Update the number of invites, if required
    if (displayInvites) then
      currentInfo.labelInvites:SetText("("..CalendarEventGetNumInvites()..")");
    end

    -- Show/Hide the 'Roles' heading
    currentInfo.labelInvites:SetShown(displayInvites);
  end

  -- Show/Hide the Class and Roles headings
  currentInfo.headingClass:SetShown(not displayRoles);
  if ((displayRoles) or (currentInfo.headingRoles ~= nil)) then
    -- Create the 'Roles' heading, if required
    if (currentInfo.headingRoles == nil) then
      currentInfo.headingRoles = currentInfo.list:CreateFontString(nil,
        "ARTWORK", "GameFontHighlightSmall");
      currentInfo.headingRoles:SetText(L["Roles"]);
      currentInfo.headingRoles:SetPoint("BOTTOMLEFT",
        currentInfo.headingClass, 0, 2);
    end

    -- Show/Hide the 'Roles' heading
    currentInfo.headingRoles:SetShown(displayRoles);
  end

  -- Set up the invite filters, if required
  if ((displayRoles) or (#currentInfo.filterBtn > 0)) then
    -- Show/Hide the filter menu button
    if (currentInfo.filterBtn[0] == nil) then
      -- Create the Filter button, if required
      -- NOTE: The button has to be named because of template requirements
      child = CreateFrame("Button", nil,
        currentInfo.list, "NigglesFilterButtonTemplate");
      child:SetWidth(80);
      child:SetText(FILTER);
      child:SetPoint("BOTTOMLEFT", currentInfo.list, "TOPLEFT", 0, 16);
      child:SetScript("OnClick", inviteFilterOnClick);
      child:SetScript("OnEvent", inviteFilterOnEvent);
      child:RegisterEvent("CALENDAR_UPDATE_INVITE_LIST");
      child.panelInfo = currentInfo;
      currentInfo.filterBtn[0] = child;
    end
    currentInfo.filterBtn[0]:SetShown(displayRoles);

    -- Show/Hide the filter buttons
    for idx = 1, #roleInfo do
      -- Create the filter buttons, if required
      if (currentInfo.filterBtn[idx] == nil) then
        child = CreateFrame("CheckButton", nil, currentInfo.list,
          "NigglesInviteListFilterTemplate");
        child:SetPoint("BOTTOMLEFT", currentInfo.list,
          "TOPLEFT", 43+(idx*40), 14);
        child.role:SetTexCoord(
          roleInfo[idx].icon[1], roleInfo[idx].icon[2],
          roleInfo[idx].icon[3], roleInfo[idx].icon[4]);
        child:SetID(idx);
        child.panelInfo = currentInfo;
        currentInfo.filterBtn[idx] = child;
      end
      currentInfo.filterBtn[idx]:SetShown(displayRoles);
      currentInfo.filterBtn[idx]:SetChecked(true);
    end

    -- Make sure the filtered bar exists
    if (currentInfo.filteredBar == nil) then
      child = CreateFrame("Button", nil, currentInfo.list,
        "NigglesFilteredBarTemplate");
      child:SetPoint("TOPLEFT", currentInfo.list, "TOPLEFT", 2, -2);
      child:SetPoint("RIGHT", currentInfo.scrollFrame.scrollChild,
        "RIGHT", -2, 0);
      child:SetScript("OnClick", inviteFileredBarOnClick);
      child.panelInfo = currentInfo;
      child:SetFrameLevel(currentInfo.scrollFrame:GetFrameLevel()+1);
      child:Show();
      currentInfo.filteredBar = child;
    end
  end

  -- Set up role/status frames for each button in the scroll frame, if required
  if ((displayOnline) or (displayRoles) or (#roleStatusFrames > 0)) then
    buttons    = currentInfo.scrollFrame.buttons;
    numButtons = #buttons;
    for buttonIdx = 1, numButtons do
      -- Initialise some variables
      button = buttons[buttonIdx];

      -- Create a role/status frame for the button, if required
      if (((displayOnline) or (displayRoles)) and
          (roleStatusFrames[buttonIdx] == nil)) then
        roleStatus = CreateFrame("Frame", nil, button,
          "NigglesCalendarRoleStatusTemplate");
        roleStatusFrames[buttonIdx] = roleStatus;
      else
        roleStatus = roleStatusFrames[buttonIdx];
      end

      -- Configure the role/status frame
      if (roleStatus ~= nil) then
        -- Change the role/status frame's parent
        roleStatus:SetParent(button);
        roleStatus:SetAllPoints();
        roleStatus.panelInfo = currentInfo;

        -- Show/Hide the button's role/status frame & its textures
        roleStatus:SetShown(displayOnline or displayRoles);
        roleStatus.onlineStatus:SetShown(displayOnline);
        for idx = 1, #roleStatus.roles do
          roleStatus.roles[idx]:SetShown(displayRoles);
        end
      end

      -- Show/Hide the button's class
      _G[button:GetName().."Class"]:SetShown(not displayRoles);
    end
  end

  -- Save the original function for updating the scroll frame, if required
  if (currentInfo.updateOrigFunc == nil) then
    currentInfo.updateOrigFunc = currentInfo.scrollFrame.update;
  end

  -- Set the function for updating the scroll frame and update it
  if ((displayOnline) or (displayRoles)) then
    _G[currentInfo.updateOrigName] = _G[currentInfo.updateName];
    currentInfo.scrollFrame.update = _G[currentInfo.updateName];
    filteredInvitesUpdate(currentInfo);
  else
    _G[currentInfo.updateOrigName] = currentInfo.updateOrigFunc;
    currentInfo.scrollFrame.update = currentInfo.updateOrigFunc;
    currentInfo.scrollFrame.update();
  end

  -- Force a guild roster update
  if (IsInGuild()) then
    GuildRoster();
  end

  return;
end

--
-- Function to update the Role/Status frames in a Calendar Invites scroll frame
--
-- NOTE: This is a merged and modified version of:
--
--         'CalendarCreateEventInviteListScrollFrame_Update()'
--         'CalendarViewEventInviteListScrollFrame_Update()'
--
--       in 'Blizzard_Calendar.lua'. It is required to allow for filtered
--       invites, based on roles.
--
local function calendarInvitesUpdate(panelInfo)
  -- Local Variables
  local button;
  local buttonHeight;
  local buttonName;
  local buttonModerator;
  local buttons;
  local characterRoles = NigglesDB.characterRoles;
  local child;
  local displayOnline = NigglesDBC.settings.CalEventsOnline;
  local displayRoles = NigglesDBC.settings.CalEventsRoles;
  local eventDay;
  local eventHour;
  local eventMin;
  local eventMonth;
  local eventYear;
  local invitee;
  local inviteIdx;
  local numInvites, namesReady = CalendarEventGetNumInvites();
  local numButtons;
  local numVisible;
  local statusIdx;
  local roleStatus;
  local scrollFrame = panelInfo.scrollFrame;
  local selectedIdx = 0;
  local editMode = -1;
  local offset;
  local inviteName;
  local className;
  local classColor;
  local classFilename;
  local inviteStatus;
  local modStatus;
  local child;
  local statusInfo;

  -- Initialise some variables
  buttons    = scrollFrame.buttons;
  numButtons = #scrollFrame.buttons;
  offset     = HybridScrollFrame_GetOffset(scrollFrame);

  -- Show/Hide the 'Retrieving...' frame
  _G[panelInfo.prefix.."FrameRetrievingFrame"]:SetShown(not namesReady);

  -- Clear any invite index
  panelInfo.panel.myInviteIndex = nil;

  -- Get the index of any selected invite
  selectedIdx = CalendarEventGetSelectedInvite();

  -- Check if the current panel is the Create Event panel
  if (panel == CalendarCreateEventFrame) then
    -- Check if the panel is in edit mode, rather than create mode
    editMode = (panel.mode == "edit" and 1 or 0);
  end

  -- Adjust the number of invites
  if (not namesReady) then
    numInvites = 0;
  elseif (filteredInvites[0] >= 0) then
    numInvites = filteredInvites[0];
  end

  -- Update the scroll frame's buttons
  numVisible = (numButtons <= offset+numInvites and numButtons or numInvites);
  for idx = 1, numVisible do
    -- Get info for the current button
    button     = buttons[idx];
    buttonName = button:GetName();
    roleStatus = roleStatusFrames[idx];

    -- Work out the calendar event invite index for the button
    if (filteredInvites[0] == -1) then
      inviteIdx = idx+offset;
    else
      inviteIdx = (filteredInvites[idx+offset] or 0);
    end

    -- Get info about the invitee
    inviteName, _, className, classFilename, inviteStatus, modStatus =
      CalendarEventGetInvite(inviteIdx);
    if (inviteName ~= nil) then
      -- Save info about the invitee
      button.inviteName = inviteName;
      button.inviteIdx  = inviteIdx;

      -- Set the moderator status for the button
      buttonModerator = _G[buttonName.."ModIcon"];
      if (modStatus == "CREATOR") then
        buttonModerator:SetTexture(
          "Interface\\GroupFrame\\UI-Group-LeaderIcon");
        buttonModerator:Show();
      elseif (modStatus == "MODERATOR") then
        buttonModerator:SetTexture(
          "Interface\\GroupFrame\\UI-Group-AssistantIcon");
        buttonModerator:Show();
      else
        buttonModerator:Hide();
      end

      -- Work out the invitee's class color
      classColor = (classFilename and RAID_CLASS_COLORS[classFilename] or
        NORMAL_FONT_COLOR);

      -- Set the name for the button
      child = _G[buttonName.."Name"];
      child:SetText(((inviteName ~= nil) and (inviteName ~= "")) and
        inviteName or UNKNOWN);
      child:SetTextColor(classColor.r, classColor.g, classColor.b);
      if (buttonModerator:IsShown()) then
        child:SetPoint("LEFT", buttonModerator, "RIGHT");
      else
        child:SetPoint("LEFT", button, "LEFT");
      end

      -- Set the class for the button
      child = _G[buttonName.."Class"];
      child:SetText(((className ~= nil) and (className ~= "")) and
        className or UNKNOWN);
      child:SetTextColor(classColor.r, classColor.g, classColor.b);

      -- Set the status for the button
      statusInfo = (inviteStatusInfo[inviteStatus] ~= nil and
         inviteStatusInfo[inviteStatus] or
         inviteStatusInfo[NIGGLES_INVITESTATUS_UNKNOWN]);
      child = _G[buttonName.."Status"];
      child:SetText(statusInfo.text);
      child:SetTextColor(statusInfo.color.r, statusInfo.color.g,
        statusInfo.color.b);

      -- (Un)select the button
      if (inviteIdx == selectedIdx) then
        CalendarCreateEventFrame_SetSelectedInvite(button);
      else
        button:UnlockHighlight();
      end

      -- Set the 'OnEnter' handler for the button, if required
      -- NOTE: This is rather inefficient. It would be better if the 'OnEvent'
      --       handler determined the edit mode and acted accordingly. This
      --       would remove any need to change handlers every time the
      --       scroll frame is updated. However it is the way Blizzard have
      --       implemented it, for whatever reason.
      if (editMode == 1) then
        button:SetScript("OnEnter", CalendarEventInviteListButton_OnEnter);
      elseif (editMode == 0) then
        button:SetScript("OnEnter", nil);
      end

      -- Check if roles are to be displayed
      if ((displayRoles) and (roleStatus ~= nil)) then
        for idx = 1, #roleStatus.roles do
          roleStatus.roles[idx]:SetTexCoord(
            roleGetTexCoord(characterRoles[inviteName], idx))
        end
      end

      -- Check if online status is to be displayed
      if ((displayOnline) and (roleStatus ~= nil)) then
        -- Work out what texture to display
        if ((UnitInRaid(inviteName)) or (UnitInParty(inviteName))) then
          statusIdx = (UnitIsConnected(inviteName) and 6 or 0);
        elseif (guildRoster[inviteName] == nil) then
          statusIdx = 5;
        elseif (bit.band(guildRoster[inviteName], 0xFFFFFFF8) ==
                 guildRoster[0]) then
          statusIdx = bit.band(guildRoster[inviteName], 0x07);
        else
          statusIdx = 0;
        end

        -- Set the texture
        roleStatus.onlineStatus:SetTexture(onlineStatusInfo[statusIdx].texture);
        roleStatus.onlineStatus:SetVertexColor(
          onlineStatusInfo[statusIdx].colour.r,
          onlineStatusInfo[statusIdx].colour.g,
          onlineStatusInfo[statusIdx].colour.b,
          onlineStatusInfo[statusIdx].colour.a);
        roleStatus.onlineStatus:SetTexCoord(
          onlineStatusInfo[statusIdx].coord[1],
          onlineStatusInfo[statusIdx].coord[2],
          onlineStatusInfo[statusIdx].coord[3],
          onlineStatusInfo[statusIdx].coord[4]);
      end

      -- Show the button
      button:Show();
    else
      -- Hide the button
      button:Hide();
    end
  end
  for idx = numVisible+1, numButtons do
    buttons[idx]:Hide();
  end

  -- Update the class buttons.
  -- NOTE: This is GROSSLY inefficient, with the entire calendar event being
  --       scanned every time the scroll frame is updated. However it is the
  --       way Blizzard have implemented it, for whatever reason.
  CalendarClassButtonContainer_Show(panelInfo.panel);

  -- Update the scroll frame
  buttonHeight = buttons[1]:GetHeight();
  HybridScrollFrame_Update(scrollFrame, numInvites*buttonHeight,
    scrollFrame:GetHeight());

  return;
end

--
-- Function to add a name to the list of character roles
--
local function characterRolesAddName(frame, name)
  -- Local Variables
  local list = Niggles.updatedOptions.characterRoles;

  -- Check the name isn't already in the list
  if (list[name] == nil) then
    -- Add the name to the list
    list[name] = 0;

    -- Update the sorted list of character role names
    characterRolesGetSortedList();

    -- Update the scroll frame
    frame.scrollFrame.update();
  end

  return;
end

--
-- Function to remove a name from the list of character roles
--
local function characterRolesDeleteName(menu, index, value)
  -- Local Variables
  local name = menu:GetData().value;

  -- Delete the name from the list
  Niggles.updatedOptions.characterRoles[name] = nil;

  -- Update the sorted list of character role names
  characterRolesGetSortedList();

  -- Update the scroll frame
  NigglesOptionsCalEvents.characterRoles.scrollFrame.update();

  return;
end

--
-- Function to check if a character has a role
--
local function characterRolesIsChecked(menu, menuArgs, index, value)
  return (bit.band(Niggles.updatedOptions.characterRoles[menuArgs[1]],
    value) == value);
end

--
-- Function to process a button click in the role sub-menus of the drop down
-- menu for the Character Roles list
--
local function characterRolesOnClick(menu, menuArgs, index, value, isChecked)
  -- Local Variables
  local characterRoles = Niggles.updatedOptions.characterRoles;
  local menuData = menu:GetData();

  -- Update the character's roles
  characterRoles[menuArgs[1]] = bit.band(characterRoles[menuArgs[1]],
    bit.rshift(0xF0, menuData.value))+(isChecked and value or 0);

  -- Update the scroll frame
  characterRolesScrollFrameUpdate();

  return;
end

--
-- Function to build an sorted list of the character names in the character
-- roles list
--
characterRolesGetSortedList = function()
  -- Create the new sorted list of character role names
  wipe(sortedCharRoles);
  for name, _ in pairs(Niggles.updatedOptions.characterRoles) do
    table.insert(sortedCharRoles, name);
  end
  table.sort(sortedCharRoles);

  return;
end

--
-- Function to get the time of the current event
--
local function currentEventGetTime()
  -- Local Variables
  local eventDay;
  local eventHour;
  local eventMin;
  local eventMonth;
  local eventYear;

  -- Update the current event time
  _, _, _, _, _, _, _, _, eventMonth, eventDay, eventYear,
    eventHour, eventMin = CalendarGetEventInfo();
  currentEventTime = NigglesMinutes(eventYear, eventMonth, eventDay,
    eventHour, eventMin);

  return;
end

--
-- Function to update the filtered list of invites
--
filteredInvitesUpdate = function(panelInfo)
  -- Local Variables
  local characterRoles = NigglesDB.characterRoles;
  local includeNoRoles;
  local inviteeRoles;
  local numInvites = CalendarEventGetNumInvites();
  local numMatches = 0;
  local roleFilter = NigglesDBC.settings.CalEventsRoleFilter;
  local roleMask = 0;

  -- Initialise some variables
  includeNoRoles = (bit.band(roleFilter, 0x8000) ~= 0);

  -- If no panel info was specified...
  if (panelInfo == nil) then
    -- ...find the first visible panel
    for idx = 1, #eventPanelInfo do
      if ((eventPanelInfo[idx].panel) and
          (eventPanelInfo[idx].panel:IsShown())) then
        panelInfo = eventPanelInfo[idx];
        break;
      end
    end
  end

  --Check a visible panel was found
  if (panelInfo ~= nil) then
    -- Check if all invites will match the filters
    if (((bitsSet[filterRoles] or 0) == #roleInfo) and
        (roleFilter == 0x8003)) then
      filteredInvites[0] = -1;
    else
      -- Create the mask used to filter all possible roles for the invites
      roleMask =
        (bit.band(roleFilter, 0x0001) ~= 0 and filterRoles or 0)+
        (bit.band(roleFilter, 0x0002) ~= 0 and bit.lshift(filterRoles, 4) or 0);

      -- Clear any existing list
      filteredInvites[0] = 0;

      -- Add all invites that match the filters or have no roles set
      for idx = 1, numInvites do
        inviteeRoles = (characterRoles[CalendarEventGetInvite(idx)] or 0);
        if (((inviteeRoles == 0) and (includeNoRoles)) or
            (bit.band(inviteeRoles, roleMask) ~= 0)) then
          numMatches = numMatches+1;
          filteredInvites[numMatches] = idx;
        end
      end
      filteredInvites[0] = numMatches;
    end

    -- Show/Hide the filtered bar
    if (panelInfo.filteredBar ~= nil) then
      if ((filteredInvites[0] == -1) or
          (filteredInvites[0] == numInvites)) then
        panelInfo.scrollFrame:SetPoint("TOPLEFT", 3, -3);
        panelInfo.filteredBar:SetHeight(1);
      else
        panelInfo.scrollFrame:SetPoint("TOPLEFT", 3, -19);
        panelInfo.filteredBar:SetFormattedText(L["FilteredShownHidden"],
          filteredInvites[0], numInvites-filteredInvites[0]);
        panelInfo.filteredBar:SetHeight(16);
      end
    end

    -- Update the scroll frame
    panelInfo.scrollFrame.update();
  end

  return;
end

--
-- Function to update the Character Roles scroll frame
--
characterRolesScrollFrameUpdate = function()
  -- Local Variables
  local button;
  local buttons;
  local character;
  local characterRoles = Niggles.updatedOptions.characterRoles;
  local isDisabled;
  local numButtons;
  local numCharacters;
  local role;
  local scrollFrame = NigglesOptionsCalEventsCharacterRolesScrollFrame;
  local topOffset;

  -- Check the scroll frame has buttons
  if ((scrollFrame.buttons ~= nil) and
      (#scrollFrame.buttons > 0)) then
    -- Initialise some variables
    buttons       = scrollFrame.buttons;
    numButtons    = #buttons;
    numCharacters = #sortedCharRoles;
    topOffset     = HybridScrollFrame_GetOffset(scrollFrame);
    isDisabled    = not buttons[1]:IsEnabled();

    -- Update the scroll frame's buttons
    for idx = 1, numButtons do
      button = buttons[idx];
      if (topOffset+idx <= numCharacters) then
        -- Set the button's character name
        character = sortedCharRoles[topOffset+idx];
        button:SetText(character);

        -- Set the character's roles
        for idx = 1, #button.roles do
          button.roles[idx]:SetTexCoord(roleGetTexCoord(
            characterRoles[character], idx))
        end

        -- Show the button
        button:Show();
      else
        button:Hide();
      end
    end

    -- Update the scroll frame's height
    HybridScrollFrame_Update(scrollFrame,
      numCharacters*buttons[1]:GetHeight(),
      numButtons*buttons[1]:GetHeight());
  else
    HybridScrollFrame_Update(scrollFrame, 0, 0);
  end

  return;
end

--
-- Function called when an event panel is shown
--
local function eventPanelOnShow()
  -- Update the guild roster info
  NigglesCalendarEventsGuildRosterUpdate();

  return;
end

--
-- Function to show/hide the Calendar Events panel, creating it if required.
--
local function eventsPanelSetShown(shown)
  -- Create the Niggles Calendar Events panel, if required
  if ((NigglesCalendarEvents == nil) and (shown)) then
    CreateFrame("FRAME", "NigglesCalendarEvents",
      UIParent, "NigglesCalendarEventsTemplate");
  end

  -- Show/hide the panel
  if (shown) then
    ShowUIPanel(NigglesCalendarEvents);
  else
    HideUIPanel(NigglesCalendarEvents);
  end

  return;
end

--
-- Function called when the Invites filter bar is clicked. It resets the filter
-- flags to a default state and updates the invite list.
--
inviteFileredBarOnClick = function(self, mouseButton)
  -- Local Variables
  local panelInfo = self.panelInfo;

  -- Reset the filter flags
  NigglesDBC.settings.CalEventsRoleFilter = 0x8001;
  filterRoles = bit.lshift(1, #roleInfo)-1;
  for idx = 1, #panelInfo.filterBtn do
    panelInfo.filterBtn[idx]:SetChecked(true);
  end

  -- Update the filtered invites
  filteredInvitesUpdate(panelInfo);

  -- Hide the dropdown menus
  CloseMenus();

  return;
end

--
-- Function to return if items should be checked in the 'Filter' menu
--
local function inviteFilterMenuOnCheck(menu, menuArgs, index, value)
  return (bit.band(NigglesDBC.settings.CalEventsRoleFilter, value) ~= 0);
end

--
-- Function to process clicks in the 'Filter' menu
--
local function inviteFilterMenuOnClick(menu, menuArgs, index, value, isChecked)
  -- Update the role filter flags
  NigglesDBC.settings.CalEventsRoleFilter =
    bit.band(NigglesDBC.settings.CalEventsRoleFilter, 0xFFFF-value)+
    (isChecked and value or 0);

  -- Update the scroll frame
  filteredInvitesUpdate();

  return true, 1;
end

--
-- Function called when the 'Filter' button is clicked in an Invite list
--
inviteFilterOnClick = function(self, menuArgs, mouseButton)
  -- Toggle the displaying of the menu
  Niggles.menu:Toggle(_L.menuInviteFilter, self);

  return;
end

--
-- Function called when a registered event fires for the Invite Filter button
--
inviteFilterOnEvent = function(self, event, ...)
  -- Process the event
  if (event == "CALENDAR_UPDATE_INVITE_LIST") then
    filteredInvitesUpdate();
  end

  return;
end

--
-- Function to work out if Prinary/Secondary Role sub-menus should be visible
-- in the 'Role/Status' menu
--
local function roleStatusMenuItemIsVisible(menu, index, value)
  return (NigglesDBC.settings.CalEventsRoles and true or false);
end

--
-- Function to work out if items should be checked in the 'Role/Status' menu
--
local function roleStatusMenuOnCheck(menu, menuArgs, index, value)
  -- Local Variables
  local isChecked = true;

  -- Work out the checked status of the menu item
  if (index == 2) then
    isChecked = (select(6, CalendarEventGetInvite(menuArgs[1])) ~= "");
  end

  return isChecked;
end

--
-- Function to process a button being clicked in the 'Role/Status' menu
--
local function roleStatusMenuOnClick(menu, menuArgs, index, value, isChecked)
  -- Local Variables
  local inviteIdx = menuArgs[1];
  local inviteName = menuArgs[2];

  -- Process the button click
  if (index == 2) then
    -- Toggle the invitee's moderator status
    local modStatus = select(6, CalendarEventGetInvite(inviteIdx));
    if (modStatus == "MODERATOR") then
      CalendarEventClearModerator(inviteIdx);
    elseif (modStatus == "") then
      CalendarEventSetModerator(inviteIdx);
    end
  elseif (index == 4) then
    StaticPopup_Show("NIGGLES_CALEVENT_REMOVE",
      CalendarEventGetInvite(inviteIdx), nil, inviteIdx);
  elseif (index == 9) then
    ChatFrame_SendTell(inviteName);
  elseif (index == 10) then
    InviteToGroup(inviteName);
  end

  return true;
end

--
-- Function to work out if items should be enabled in the 'Role/Status' menu
--
local function roleStatusMenuOnEnable(menu, menuArgs, index, value)
  -- Local Variables
  local inviteIdx = menuArgs[1];
  local isEnabled = true;
  local modStatus;

  -- Work out if the menu item should be enabled
  if ((index == 2) or (index == 4)) then
    modStatus = select(6, CalendarEventGetInvite(inviteIdx));
    isEnabled = ((CalendarEventIsModerator()) and (modStatus ~= "CREATOR"));
  end

  return isEnabled;
end

--
-- Function to work out if items should be checked in one of the role
-- sub-menus of the 'Role/Status' menu
--
local function roleStatusRoleMenuOnCheck(menu, menuArg, index, value)
  -- Local Variables
  local inviteName = menuArg[2];
  local inviteRoles;

  -- Get the invite's current roles
  inviteRoles = (NigglesDB.characterRoles[inviteName] or 0);

  -- Check if the character has the specified role
  return ((bit.band(inviteRoles, value) ~= 0) or
          ((value == 0) and (bit.band(inviteRoles,
             bit.lshift(0x0F, menu:GetData().value)) == 0)));
end

--
-- Function to process a button being clicked in one of the role sub-menus of
-- the 'Role/Status' menu
--
local function roleStatusRoleMenuOnClick(menu, menuArgs, index, value, isChecked)
  -- Local Variables
  local bitShift = menu:GetData().value;
  local inviteName = menuArgs[2];

  -- Update the character's role
  NigglesDB.characterRoles[inviteName] =
    bit.band((NigglesDB.characterRoles[inviteName] or 0),
      bit.rshift(0xF0, bitShift))+(isChecked and value or 0);

  -- Update the scroll frame
  filteredInvitesUpdate();

  return true;
end

--
-- Function to work out if items should be checked in the 'Invite Status'
-- sub-menu of the 'Role/Status' menu
--
local function roleStatusInviteMenuOnCheck(menu, menuArgs, index, value)
  -- Check if the menu item is for the character's current invite state
  return (value == select(5, CalendarEventGetInvite(menuArgs[1])));
end

--
-- Function to process the click on an item in the 'Invite Status' sub-menu of
-- the 'Role/Status' menu
--
local function roleStatusInviteMenuOnClick(menu, menuArgs, index, value)
  -- Update the invite status, if required
  local name, _, _, _, inviteStatus = CalendarEventGetInvite(menuArgs[1]);
  if ((value ~= inviteStatus) and (name == menuArgs[2])) then
    CalendarEventSetStatus(menuArgs[1], value);
  end

  return;
end

--
-- Function to work out if items should be enabled in the 'Invite Status'
-- sub-menu of the 'Role/Status' menu
--
local function roleStatusInviteMenuOnEnable(menu, menuArgs, index, value)
  -- Check if the menu item is for a valid invite status
  return (bit.band(menuArgs[3], bit.lshift(1, value-1)) > 0)
end

--
-- Function to create a table containing the items for a 'Role' sub-menu
--
function roleSubMenuGetItems(menu, menuArgs, value)
  -- Local Variables
  local items = {};

  -- Add the invite status menu items to the 'RoleStatus' menu data
  for roleIdx = 0, #roleInfo do
    items[roleIdx+1] =
    {
      type  = "radioButton",
      label = roleInfo[roleIdx].text,
      color = roleInfo[roleIdx].colorCode,
      value = bit.lshift(roleInfo[roleIdx].flag, value);
    };
  end

  return items;
end

--
-- Function to return the texture coordinates for a given role flag
--
roleGetTexCoord = function(roles, index)
  -- Local Variables
  local roleIdx = 0;
  local role = bit.band(bit.rshift((roles or 0), (index-1)*4), 0x0F);

  -- Search for the flag in the list of roles
  for idx = 0, #roleInfo do
    if (roleInfo[idx].flag == role) then
      roleIdx = idx;
      break;
    end
  end

  -- Return the role's texture coordinates
  return roleInfo[roleIdx].icon[1], roleInfo[roleIdx].icon[2],
         roleInfo[roleIdx].icon[3], roleInfo[roleIdx].icon[4];
end

-------------------------------------------------------------------------------
--                               M  E  N  U  S
-------------------------------------------------------------------------------

_L.menuCharacterRoles =
{
  style    = "menu",
  position = "cursor",
  offset   = {x = 2, y = 2},
  value    = "",
  items    =
  {
    {type = "label", label = "|1"},
    {
      type    = "submenu",
      label   = L["RolePrimary"],
      hideAll = true;
      onCheck = characterRolesIsChecked,
      onClick = characterRolesOnClick,
      items   = roleSubMenuGetItems,
      value = 0,
    },
    {
      type    = "submenu",
      label   = L["RoleSecondary"],
      hideAll = true;
      onCheck = characterRolesIsChecked,
      onClick = characterRolesOnClick,
      items   = roleSubMenuGetItems,
      value   = 4,
    },
    {
      type    = "button",
      label   = DELETE,
      onClick = characterRolesDeleteName,
    }
  }
};

_L.menuInviteFilter =
{
  style    = "menu",
  position = "right",
  offset   = {x = -18, y = -8},
  onCheck  = inviteFilterMenuOnCheck,
  onClick  = inviteFilterMenuOnClick,
  items    =
  {
    {type = "checkButton", label = L["RolePrimary"],   value = 0x0001},
    {type = "checkButton", label = L["RoleSecondary"], value = 0x0002},
    {type = "checkButton", label = L["No Roles"],      value = 0x8000},
  }
};

_L.menuRoleStatus =
{
  style    = "menu",
  position = "cursor",
  offset   = {x = 2, y = 2},
  value    = "",
  onCheck  = roleStatusMenuOnCheck,
  onClick  = roleStatusMenuOnClick,
  onEnable = roleStatusMenuOnEnable,
  items    =
  {
    {type = "label", label = "|2"},
    {type = "checkButton", label = L["Moderator"]},
    {
      type     = "subMenu",
      label    = L["Invite Status"],
      hideAll  = true;
      onCheck  = roleStatusInviteMenuOnCheck,
      onClick  = roleStatusInviteMenuOnClick,
      onEnable = roleStatusInviteMenuOnEnable,
      items    = {},
    },
    {type = "button", label = REMOVE},
    {type = "divider"},
    {
      type      = "subMenu",
      label     = L["Primary Role"],
      isVisible = roleStatusMenuItemIsVisible,
      hideAll   = true;
      onCheck   = roleStatusRoleMenuOnCheck,
      onClick   = roleStatusRoleMenuOnClick,
      items     = roleSubMenuGetItems,
      value     = 0,
    },
    {
      type      = "subMenu",
      label     = L["Secondary Role"],
      isVisible = roleStatusMenuItemIsVisible,
      hideAll   = true;
      onCheck   = roleStatusRoleMenuOnCheck,
      onClick   = roleStatusRoleMenuOnClick,
      items     = roleSubMenuGetItems,
      value     = 4,
    },
    {type = "divider"},
    {type = "button", label = L["Whisper"]},
    {type = "button", label = L["Invite to Group"]}
  },
};

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to enable/disable
--
function NigglesCalendarEnable()
  -- Hide the 'Calendar Events' frame, if required
  if (NigglesDBC.settings.CalEventsList ~= true) then
    HideUIPanel(NigglesCalendarEvents);
  end

  -- Show/Hide the Size Down button in the Calendar panel, if required
  if (NigglesCalendarSizeDownButton ~= nil) then
    NigglesCalendarSizeDownButton:SetShown(NigglesDBC.settings.CalEventsList);
  end

  return;
end

--
-- Function to show/hide the Calendar Events panel
--
function NigglesCalendarEventsOnClick(self, button)
  -- Check the 'Calendar Events' frame is enabled
  if (NigglesDBC.settings.CalEventsList == true) then
    --Show/Hide the Calendar and Calendar Events panels
    if (button == "LeftButton") then
      -- Hide the Calendar Events panel
      eventsPanelSetShown(false);
    elseif (button == "RightButton") then
      -- Hide the Calendar frame
      Calendar_Hide();

      -- Toggle the Calendar Events panel
      eventsPanelSetShown((NigglesCalendarEvents == nil) or
        (not NigglesCalendarEvents:IsVisible()));
    end
  end

  return;
end

--
-- Function to initialise the Calendar Events panel
--
function NigglesCalendarEventsOnLoad(self)
  -- Local Variables
  local currentInfo;
  local menuItems;

  -- Work out which PvP logo to display
  if (UnitFactionGroup("player") == "Alliance") then
    calendarEventTypeTexture[CALENDAR_EVENTTYPE_PVP].default =
      calendarEventTypeTexture[CALENDAR_EVENTTYPE_PVP].default.."02";
  else
    calendarEventTypeTexture[CALENDAR_EVENTTYPE_PVP].default =
      calendarEventTypeTexture[CALENDAR_EVENTTYPE_PVP].default.."01";
  end

  -- Set the attributes that make the panel a UI Panel
  self:SetAttribute("UIPanelLayout-defined",   true);
  self:SetAttribute("UIPanelLayout-enabled",   true);
  self:SetAttribute("UIPanelLayout-area",      "left");
  self:SetAttribute("UIPanelLayout-pushable",  2);
  self:SetAttribute("UIPanelLayout-whileDead", true);
  self:SetAttribute("UIPanelLayout-width",     self:GetWidth()+65);

  -- Hide the panel's button bar
  ButtonFrameTemplate_HideButtonBar(self);

  -- Set the panel's title
  self.TitleText:SetText("Calendar Events");

  -- Set the panel's portrait
  self.portrait:SetTexture(
     "Interface\\AddOns\\Niggles\\textures\\calendar");

  -- Adjust the size/position of the panel's inset
  self.Inset:SetPoint("TOPLEFT", 4, -65);
  self.Inset:SetPoint("BOTTOMRIGHT", -7, 6);

  -- Initialise the panel's ownership info for the Calendar children
  ownershipInfo[self] = {};
  currentInfo = ownershipInfo[self];
  for childIdx, childInfo in ipairs(ownershipChildren) do
    -- Initialise the child's info
    currentInfo[childIdx]        = {};
    currentInfo[childIdx].frame  = _G[childInfo.name];
    currentInfo[childIdx].points = {};
    if ((childInfo.x ~= nil) and (childInfo.y ~= nil)) then
      currentInfo[childIdx].points[1] =
        {"TOPLEFT", self, "TOPLEFT", childInfo.x, childInfo.y};
    end
  end

  -- Initialise the scroll frame
  self.scrollFrame.update = calendarEventsUpdate;
  HybridScrollFrame_CreateButtons(self.scrollFrame,
    "NigglesCalendarEventButtonTemplate", 1, 0);
  self.scrollFrame.scrollBar.doNotHide = true;

  -- Register for calendar update events
  self:RegisterEvent("CALENDAR_NEW_EVENT");
  self:RegisterEvent("CALENDAR_OPEN_EVENT");
  self:RegisterEvent("CALENDAR_UPDATE_PENDING_INVITES");
  self:RegisterEvent("CALENDAR_UPDATE_ERROR");
  self:RegisterEvent("CALENDAR_UPDATE_EVENT_LIST");

  return;
end

--
-- Function to process events for the Calendar Events panel
--
function NigglesCalendarEventsOnEvent(frame, event, ...)
  -- Local Variables
  local calendarMonth;
  local eventType;
  local calendarYear;
  local currentMonth;
  local currentYear;

  -- Check that the Calendar Events panel is visible
  if (NigglesCalendarEvents:IsVisible()) then
    -- Process the event
    if (event == "CALENDAR_OPEN_EVENT") then
      -- Check the calendar event type should be processed
      eventType = ...;
      if ((eventType == "PLAYER") or
          (eventType == "GUILD_EVENT") or
          (eventType == "GUILD_ANNOUNCEMENT")) then
        -- Check if the calendar event can be edited
        if (CalendarEventCanEdit()) then
          CalendarCreateEventFrame.mode = "edit";
          CalendarFrame_ShowEventFrame(CalendarCreateEventFrame);
        else
          CalendarFrame_ShowEventFrame(CalendarViewEventFrame);
        end
      end
    else
      -- Get the current date
      _, currentMonth, _, currentYear = CalendarGetDate();

      -- Get the month displayed by the calendar
      calendarMonth, calendarYear = CalendarGetMonth();

      -- Check the calendar is displaying the current month
      if ((calendarMonth == currentMonth) or
          (calendarYear  == currentYear)) then
        calendarEventsGetList();
      else
        CalendarSetAbsMonth(currentMonth, currentYear);
      end
    end
  end

  return;
end

--
-- Function to process events for the Calendar Events panel
--
function NigglesCalendarEventsOnShow(self)
  -- Local Variables
  local month;
  local year;

  -- Take ownership of the Calendar's children
  calendarEventsTakeOwnership(self);

  -- Make sure the calendar is set for the current month
  _, month, _, year = CalendarGetDate();
  CalendarSetAbsMonth(month, year);

  -- Update the list of events
  calendarEventsGetList();

  return;
end

--
-- Function to process clicks on Calendar Event buttons
--
function NigglesCalendarEventButtonOnClick(button)
  -- Check there is a calendar event associated with the button
  if (button.event ~= nil) then
    -- Open the calendar event
    CalendarOpenEvent(button.event.monthOffset, button.event.day,
      button.event.index);
  end

  return;
end

--
-- Function called when the guild roster is updated
--
function NigglesCalendarEventsGuildRosterUpdate()
  -- Local Variables
  local numMembers = GetNumGuildMembers();
  local name;
  local isMobile;
  local isOnline;
  local status;

  -- Check one of the frames that needs the guild roster is visible
  if ((NigglesDBC.settings.CalEventsOnline == true) and
      ((CalendarCreateEventFrame ~= nil) and
       (CalendarCreateEventFrame:IsShown())) or
      ((CalendarViewEventFrame ~= nil) and
       (CalendarViewEventFrame:IsShown()))) then
    -- Increment the guild roster counter
    guildRoster[0] = guildRoster[0]+8;

    -- Update the guild roster
    for idx = 1, numMembers do
      name, _, _, _, _, _, _, _, isOnline, status, _, _, _, isMobile =
        GetGuildRosterInfo(idx);
      if (name ~= nil) then
        name = Ambiguate(name, "none");
        if (isOnline) then
          guildRoster[name] = guildRoster[0]+1;
        elseif (isMobile) then
          guildRoster[name] = guildRoster[0]+2+status;
        else
          guildRoster[name] = guildRoster[0];
        end
      end
    end

    -- Make sure the player is in the guild roster, even if not in a guild
    guildRoster[UnitName("PLAYER")] = guildRoster[0]+1;

    -- Update any Calendar Event frame currently shown
    for idx = 1, #eventPanelInfo do
      if ((eventPanelInfo[idx].panel) and
          (eventPanelInfo[idx].panel:IsShown())) then
        filteredInvitesUpdate(eventPanelInfo[idx]);
      end
    end
  end

  return;
end

--
-- Function called when Blizzard's Calendar addon is loaded
--
function NigglesCalendarOnLoad()
  -- Local Variables
  local numPoints;
  local currentInfo;
  local button;
  local itemData;

  -- Add a hook to update the event frames as they are show
  hooksecurefunc("CalendarFrame_ShowEventFrame",
    calendarInvitesShow);

  -- Initialise the ownership info for the Calendar children
  ownershipInfo[CalendarFrame] = {};
  currentInfo = ownershipInfo[CalendarFrame];
  for childIdx, childInfo in ipairs(ownershipChildren) do
    -- Initialise the child's info
    currentInfo[childIdx]        = {};
    currentInfo[childIdx].frame  = _G[childInfo.name];
    currentInfo[childIdx].points = {};

    -- Initialise the child's anchor points
    numPoints = _G[childInfo.name]:GetNumPoints();
    for pointIdx = 1, numPoints do
      currentInfo[childIdx].points[pointIdx] =
        {_G[childInfo.name]:GetPoint(pointIdx)};
    end
  end

  -- Hook into the Calendar frames
  CalendarFrame:HookScript("OnShow", calendarEventsTakeOwnership);

  -- Initialise info for the event panels
  for panelIdx = 1, #eventPanelInfo do
    -- Initialise some variables
    currentInfo = eventPanelInfo[panelIdx];

    -- Get the event panel
    currentInfo.panel = _G[currentInfo.prefix.."Frame"];

    -- Set the event panel's ID
    currentInfo.panel:SetID(panelIdx);

    -- Hook into the event panel
    currentInfo.panel:HookScript("OnShow", eventPanelOnShow);

    -- Get handles for some of the panel's children
    currentInfo.divider        =
      _G[currentInfo.prefix.."Divider"];
    currentInfo.list           =
      _G[currentInfo.prefix.."InviteList"];
    currentInfo.scrollFrame    =
      _G[currentInfo.prefix.."InviteListScrollFrame"];
    currentInfo.headingClass   =
      _G[currentInfo.prefix.."InviteListClassSortButton"];

    -- Adjust the invite list's layout
    currentInfo.list:SetPoint("BOTTOM", 0, currentInfo.yOffset);

    -- Save the scroll frame and scroll bar points
    NigglesFramePointsGet(currentInfo.scrollFrame,
      currentInfo.scrollPoints[1]);
    NigglesFramePointsGet(currentInfo.scrollFrame.scrollBar,
      currentInfo.scrollPoints[2]);
  end

  -- Move the Filter drop down
  CalendarFilterFrame:SetPoint("TOPRIGHT", -80, -21);

  -- Create a Size Down Button for the Calendar panel
  button = CreateFrame("BUTTON", "NigglesCalendarSizeDownButton",
    CalendarFrame, "NigglesCalendarSizeButtonTemplate");
  button:SetShown(NigglesDBC.settings.CalEventsList);

  -- Add the invite status menu items to the 'RoleStatus' menu data
  menuItems = _L.menuRoleStatus.items[3].items;
  for _, statusIdx in ipairs(inviteStatusOrder) do
    tinsert(menuItems,
      {
        type  = "radioButton",
        label = inviteStatusInfo[statusIdx].text,
        color = inviteStatusInfo[statusIdx].colorCode,
        value = statusIdx
      });
  end

  return;
end

--
-- Function called when a role/status frame is clicked
--
function NigglesCalendarRoleStatusOnClick(self, mouseButton)
  -- Local Variables
  local inviteIdx = self:GetParent().inviteIdx;
  local validStatus = 0;
  local currentStatus;

  -- Check if the menu needs to be shown
  if (not Niggles.menu:IsShown(self:GetParent())) then
    -- Create a mask for the valid invite status
    if (CalendarEventIsModerator()) then
      validStatus = bit.lshift(1,
        select(5, CalendarEventGetInvite(inviteIdx))-1);
      for _, status in ipairs({CalendarEventGetStatusOptions(inviteIdx)}) do
        if (type(status) == "number") then
          validStatus = validStatus+bit.lshift(1, status-1);
        end
      end
    end
    Niggles.menu:Show(_L.menuRoleStatus, self:GetParent(),
      inviteIdx, CalendarEventGetInvite(inviteIdx), validStatus);
  else
    Niggles.menu:Hide();
  end

  return;
end

--
-- Function called when the mouse enters a role/status frame
--
function NigglesCalendarRoleStatusOnEnter(self)
  -- Local Variables
  local parent = self:GetParent();

  -- Display a tooltip for the frame's parent
  CalendarEventInviteListButton_OnEnter(parent, motion);

  -- Lock the highlight for the frame's parent
  parent:LockHighlight();

  return;
end

--
-- Function called when the mouse leaves a role/status frame
--
function NigglesCalendarRoleStatusOnLeave(self)
  -- Hide the game tooltip
   GameTooltip:Hide();

  -- Unlock the highlight for the frame's parent
  self:GetParent():UnlockHighlight();

  return;
end

--
-- Function called when a role/status frame is loaded
--
function NigglesCalendarRoleStatusOnLoad(self)
  -- Local Variables
  local texture;

  -- Create the textures for the roles
  self.roles = {};
  for idx = 1, #rolePref do
    texture = self:CreateTexture(nil, "ARTWORK");
    texture:SetTexture("Interface\\LFGFRAME\\UI-LFG-ICON-PORTRAITROLES");
    texture:SetWidth(14);
    texture:SetHeight(14);
    texture:SetPoint("CENTER", -27+(idx*15), 0);
    self.roles[idx] = texture;
  end

  return;
end

--
-- Function to process updates for the current calendar event
--
function NigglesCalEventOnUpdate()
  -- Update any shown event panel
  for idx = 1, #eventPanelInfo do
    -- Check if the event panel is shown
    if ((eventPanelInfo[idx].panel ~= nil) and
        (eventPanelInfo[idx].panel:IsShown())) then
      -- Update the time for the current event
      currentEventGetTime();

      -- Update the number of invites, if required
      if (NigglesDBC.settings.CalEventsInvites) then
        eventPanelInfo[idx].labelInvites:SetText("("..
          CalendarEventGetNumInvites()..")");
      end

      -- Update the invite list
      eventPanelInfo[idx].scrollFrame.update();
    end
  end

  return;
end

--
-- Function to update the scroll frame for the Invites list of the
-- 'Create Event' panel
--
function NigglesCreateEventInviteListUpdate()
  calendarInvitesUpdate(eventPanelInfo[1]);
  return;
end;

--
-- Function called when an Invites filter check button is clicked. It
-- updates the filter flags and updates the invite list
--
function NigglesInvitesFilterOnClick(self)
  -- Local Variables
  local roleIdx = self:GetID();

  -- Check there is some info for the check button's role
  if (roleInfo[roleIdx] ~= nil) then
    -- Update the filter flags
    filterRoles = bit.band(filterRoles, bit.bnot(roleInfo[roleIdx].flag))+
      (self:GetChecked() and roleInfo[roleIdx].flag or 0);

    -- Update the filtered invites
    filteredInvitesUpdate(self.panelInfo);
  end

  return;
end

--
-- Function called when the mouse enters an Invites filter check button. It
-- displays the tooltip for the check button.
--
function NigglesInvitesFilterOnEnter(self)
  -- Local Variables
  local roleIdx = self:GetID();

  -- Check there is some info for the check button's role
  if (roleInfo[roleIdx] ~= nil) then
    GameTooltip:SetOwner(self, "ANCHOR_LEFT");
    GameTooltip:SetText(string.match(roleInfo[roleIdx].text, "^|T.*|t(.\+)$"),
      nil, nil, nil, nil, 1);
    GameTooltip:Show();
  end

  return;
end

--
-- Function to initialise the Options panel for Calendar Events
--
function NigglesOptionsCalEventsInit(object, setting)
  -- Initialise the specified object
  if (object == NigglesOptionsCalEventsCharacterRoles) then
    characterRolesGetSortedList();
    NigglesCharacterListInit(object, L["Character Roles"],
      "NigglesCharacterRoleBtnTemplate",
      characterRolesScrollFrameUpdate, characterRolesAddName);
  end

  return;
end

--
-- Function called when a 'Character Roles' button is loaded
--
function NigglesCharacterRolesBtnOnLoad(self)
  -- Local Variables
  local texture;

  -- Register for left and right clicks
  self:RegisterForClicks("LeftButtonUp", "RightButtonUp");

  -- Clear the pushed offset
  self:SetPushedTextOffset(0, 0);

  -- Create the textures for the roles
  self.roles = {};
  for idx = 1, #rolePref do
    texture = self:CreateTexture(nil, "ARTWORK");
    texture:SetTexture("Interface\\LFGFRAME\\UI-LFG-ICON-PORTRAITROLES");
    texture:SetWidth(14);
    texture:SetHeight(14);
    texture:SetPoint("TOPRIGHT", -15*(#rolePref-idx), -1);
    self.roles[idx] = texture;
  end

  return;
end

--
-- Function called when a 'Character Roles' button is clicked
--
function NigglesCharacterRolesBtnOnClick(self, mouseButton)
  -- Get the name of the character the button is for
  local name = self:GetText();

  -- Show/Hide the menu
  Niggles.menu:Toggle(_L.menuCharacterRoles, self:GetParent(), name);

  return;
end

--
-- Function called when a 'Character Roles' button is enabled or disabled
--
function NigglesCharacterRolesBtnOnEnable(self)
  -- Local Variables
  local isDisabled = not self:IsEnabled();

  -- Enable/Disable the button's textures
  for idx = 1, #self.roles do
    self.roles[idx]:SetDesaturated(isDisabled);
  end

  return;
end

--
-- Function to update the scroll frame for the Invites list of the
-- 'View Event' panel
--
function NigglesViewEventInviteListUpdate()
  calendarInvitesUpdate(eventPanelInfo[2]);
  return;
end

--
-- Function to toggle between the Calendar and Events panels
--
function NigglesCalendarViewToggle(self, button)
  -- Switch between the Calendar and Events panels
  if (self:GetParent() == CalendarFrame) then
    Calendar_Hide();
    eventsPanelSetShown(true);
  elseif (self:GetParent() == NigglesCalendarEvents) then
    eventsPanelSetShown(false);
    Calendar_Show();
  end

  return;
end
