-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local autoLootSlots = {};

local stackSplitMax;
local stackSplitMaxPurchasable;

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local stackSplitMaxOnClick;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to auto-loot items with a short delay between each
--
local function autoLootNext()
  -- Local Variables
  local slotIdx;
  local isLocked;
  local lootIdx;

  -- Search for an item that can be looted
  while ((next(autoLootSlots)) and (lootIdx == nil)) do
    slotIdx = autoLootSlots[#autoLootSlots];
    _, _, _, _, isLocked = GetLootSlotInfo(slotIdx);
    if (isLocked == false) then
      lootIdx = slotIdx;
    end
    autoLootSlots[#autoLootSlots] = nil;
  end

  -- If a suitable item was found, loot it
  if (lootIdx ~= nil) then
    LootSlot(lootIdx);
  end

  -- Loot another item after a short delay, if required
  if (next(autoLootSlots)) then
    C_Timer.After(LOOTFRAME_AUTOLOOT_DELAY, autoLootNext);
  end

  return;
end

--
-- Function to delete an item in a bag or bank slot with a modified click.
-- The deletion is confirmed by calling the appropriate Blizzard function.
--
local function itemDelete(bagId, slotId)
  -- Check the option is enabled
  if (NigglesDBC.settings.UiItemDelete == true) then
    -- Get info for the item clicked on
    local itemId = GetContainerItemID(bagId, slotId);

    -- Check the click isn't already handled
    if ((itemId) and (IsAltKeyDown())) then
      -- Pick up the item clicked on
      PickupContainerItem(bagId, slotId);

      -- Get the quality of the item clicked on
      local itemName, _, itemQuality = GetItemInfo(itemId);

      -- Trigger the standard delete item action
      UIParent_OnEvent(self, "DELETE_ITEM_CONFIRM", itemName, itemQuality);
    end
  end

  return;
end

--
-- Function to delete an item in a bag with a modified click.
--
local function bagItemClicked(self, button)
  -- Delete the item
  itemDelete(self:GetParent():GetID(), self:GetID());

  return;
end

--
-- Function to delete an item in a bank slot with a modified click.
--
local function bankItemClicked(self, button)
  -- Delete the item
  itemDelete(BANK_CONTAINER, self:GetID());

  return;
end

--
-- Function to zoom the minimap all the way in when the button is
-- right-clicked.
--
local function minimapZoomIn(self, button)
  -- Check the option is enabled
  if (NigglesDBC.settings.UiMiniMapZoom == true) then
    -- Check the right button was clicked and the 'Zoom In' button is enabled
    if ((button == "RightButton") and
        (MinimapZoomIn:IsEnabled())) then
      MinimapZoomIn:Disable();
      MinimapZoomOut:Enable();
      PlaySound("igMinimapZoomIn");
      Minimap:SetZoom(Minimap:GetZoomLevels()-1);
    end
  end

  return;
end

--
-- Function to zoom the minimap all the way out when the button is
-- right-clicked.
--
local function minimapZoomOut(self, button)
  -- Check the option is enabled
  if (NigglesDBC.settings.UiMiniMapZoom == true) then
    -- Check the right button was clicked and the 'Zoom Out' button is enabled
    if ((button == "RightButton") and
        (MinimapZoomOut:IsEnabled())) then
      MinimapZoomIn:Enable();
      MinimapZoomOut:Disable();
      PlaySound("igMinimapZoomOut");
      Minimap:SetZoom(0);
    end
  end

  return;
end

--
-- Function to toggle Equipment Manager Ignore for a item slot.
-- This function implements functionality from the function
-- 'PaperDollFrameItemFlyoutButton_OnClick' in PaperDollFrame.lua.
--
local function paperDollItemSlotClicked(self)
  -- Check the option is enabled
  if (NigglesDBC.settings.UiEMIgnoreToggle == true) then
    -- Check the correct modifier key is down and Equipment Manager is shown
    if ((IsAltKeyDown()) and
        ((PaperDollEquipmentManagerPane:IsShown() and
         (PaperDollEquipmentManagerPane.selectedSetName)) or
         (GearManagerDialogPopup:IsShown())) and
         (self.popoutButton:IsShown())) then
      -- Get the ID of the slot
      local slotId = self:GetID();

      -- Toggle the ignoring of the slot
      if (EquipmentManagerIsSlotIgnoredForSave(slotId)) then
        EquipmentManagerUnignoreSlotForSave(slotId);
        self.ignored = nil;
      else
        EquipmentManagerIgnoreSlotForSave(slotId);
        self.ignored = true;
      end

      -- Update the slot and the flyout menu
      PaperDollItemSlotButton_Update(self);
      EquipmentFlyout_Show(self);
      PaperDollEquipmentManagerPaneSaveSet:Enable();
    end
  end

  return;
end

--
-- Function to process clicking of Niggles Stack Split 'Maximum' button
--
stackSplitMaxOnClick = function(self, button)
  -- Hide the 'Split Stack' frame
  StackSplitFrame:Hide();

  -- Check the 'Split Stack' frame has an owner
  if (StackSplitFrame.owner) then
    -- Call the owner's 'Split Stack' function, with max stack size
    StackSplitFrame.owner.SplitStack(StackSplitFrame.owner,
      stackSplitMaxPurchasable);
  end

  return;
end

--
-- Function called when the 'OpenStackSplit' function is called. It stores the
-- maximum amount purchasable and displays the 'Maximum' frame.
--
local function stackSplitOpen(maxPurchasable, parent, anchor, anchorTo)
  -- Local Variables
  local owner;
  local level;

  -- Save the maximum purchasable
  stackSplitMaxPurchasable = maxPurchasable;

  -- Find the ultimate owner of the 'Stack Split' frame
  owner = StackSplitFrame.owner;
  while ((owner ~= nil) and
         (owner ~= MerchantFrame) and
         (owner ~= UIParent)) do
    owner = owner:GetParent();
  end

  -- Show/Hide the 'Stack Split Max' frame, depending on the owner
  if ((NigglesDBC.settings.UiMerchantBuyMax == true) and
      (owner == MerchantFrame) and
      (maxPurchasable > 1)) then
    -- Create the 'Stack Split Max' frame, if required
    if (stackSplitMax == nil) then
      -- Create the 'Stack Split Max' frame
      stackSplitMax = CreateFrame("Frame", nil, StackSplitFrame,
        "NigglesStackSplitMaxTemplate");
      stackSplitMax.button:SetScript("OnClick", stackSplitMaxOnClick);

      -- Adjust the level of the 'Stack Split' buttons
      level = stackSplitMax:GetFrameLevel()+1;
      StackSplitOkayButton:SetFrameLevel(level);
      StackSplitCancelButton:SetFrameLevel(level);
    end

    -- Initialise the 'Stack Split Max' frame and show it
    stackSplitMax.button:SetText(format("%s (%d)", MAXIMUM, maxPurchasable));
    stackSplitMax:Show();
  else
    -- Hide the 'Stack Split Max' frame, if it exists
    if (stackSplitMax ~= nil) then
      stackSplitMax:Hide();
    end
  end

  return;
end

--
-- Function called when a tooltip is set with an aura. It adds the caster of
-- the buff to the tooltip.
--
local function tooltipSetUnitAura(tooltip, unit, index, filter)
  -- Check if the tooltip should be updated
  if ((NigglesDBC.settings.UiBuffCaster) and
      (unit == "player") and
      (filter == "HELPFUL")) then
    -- Check the tooltip is for a buff
    local owner = GameTooltip:GetOwner();
    if ((owner ~= nil) and (owner:GetParent() == BuffFrame)) then
      -- Get the buff's caster
      local _, _, _, _, _, _, _, caster = UnitAura("player", index, "HELPFUL");
      if (caster ~= nil) then
        -- Add the buff's caster to the tooltip
        GameTooltip:AddLine(format(L["BuffCastBy"], UnitName(caster)));
        GameTooltip:Show();
      end
    end
  end

  return;
end

--
-- Function called when a tooltip is set with a consolidated buff. It adds the
-- caster of the buff to the tooltip.
--
local function tooltipSetUnitConsolidatedBuff(tooltip, unit, index)
  -- Check if the tooltip should be updated
  if ((NigglesDBC.settings.UiBuffCaster) and (unit == "player")) then
    -- Check the tooltip is for a consolidated buff
    local owner = GameTooltip:GetOwner();
    if ((owner ~= nil) and
        (owner:GetParent() == ConsolidatedBuffsTooltip)) then
      -- Get the buff's caster
      local _, _, _, _, _, _, _, caster = UnitAura("player",
        GetRaidBuffTrayAuraInfo(index));
      if (caster ~= nil) then
        -- Add the buff's caster to the tooltip
        GameTooltip:AddLine(format(L["BuffCastBy"], UnitName(caster)));
        GameTooltip:Show();
      end
    end
  end

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to automatically loot certain items
--
function NigglesAutoLoot()
  -- Local Variables
  local numLootItems = GetNumLootItems();
  local isLocked;
  local itemLink;
  local itemType;
  local settings = NigglesDBC.settings;

  -- Check if the loot is from fishing
  if ((settings.UiAutoLootFish == true) and (IsFishingLoot())) then
    for slotIdx = numLootItems, 1, -1 do
      _, _, _, _, isLocked = GetLootSlotInfo(slotIdx);
      if (isLocked == false) then
        tinsert(autoLootSlots, slotIdx);
      end
    end
  else
    -- Search for an item to auto-loot
    for slotIdx = numLootItems, 1, -1 do
      -- Check if the slot is locked
      _, _, _, _, isLocked = GetLootSlotInfo(slotIdx);
      if (isLocked == false) then
        -- Get info about the loot in the current slot
        itemType = GetLootSlotType(slotIdx);
        itemLink = GetLootSlotLink(slotIdx);

        -- Work out if the slot should be looted
        if (((itemType == LOOT_SLOT_MONEY) and
             (settings.UiAutoLootMoney)) or
            ((NigglesArchaeologyIsLoot(itemType, itemLink)) and
             (settings.UiAutoLootMoney)) or
            ((itemType == LOOT_SLOT_CURRENCY) and
             (settings.UiAutoLootCurrency))) then
          tinsert(autoLootSlots, slotIdx);
        end
      end
    end
  end

  -- Check a slot was found that should be looted
  if (next(autoLootSlots)) then
    C_Timer.After(LOOTFRAME_AUTOLOOT_DELAY, autoLootNext);
  end

  return;
end

--
-- Function to toggle all the class colour checkboxes in the Chat 'General
-- Config' panel
--
function NigglesChatConfigOnClick(button)
  -- Local variables
  local checkBox;
  local checked = false;
  local checkIdx = 1;
  local frame = button:GetParent();
  local frameName = frame:GetName();

  -- Work out what state to set the checkboxes to
  checkBox = _G[frameName.."CheckBox"..checkIdx.."ColorClasses"];
  while (checkBox ~= nil) do
    -- Check if the checkbox is visible
    if ((checkBox:IsVisible()) and
        (not checkBox:GetChecked())) then
      checked = true;
      break;
    end

    -- Move on to the next checkbox
    checkIdx = checkIdx+1;
    checkBox = _G[frameName.."CheckBox"..checkIdx.."ColorClasses"];
  end

  -- Process each of the class color checkboxes
  checkIdx = 1;
  checkBox = _G[frameName.."CheckBox"..checkIdx.."ColorClasses"];
  while (checkBox ~= nil) do
    -- Check if the checkbox is visible
    if (checkBox:IsVisible()) then
      checkBox:func(checked);
      checkBox:SetChecked(checked);
    end

    -- Move on to the next checkbox
    checkIdx = checkIdx+1;
    checkBox = _G[frameName.."CheckBox"..checkIdx.."ColorClasses"];
  end

  return;
end

--
-- Function to switch to the first/last page of a frame that uses a page
-- concept when a 'Prev' or 'Next' button is right-clicked.
--
function NigglesFirstLastPage(self, button)
  -- Check the option is enabled and the right button was clicked
  if ((NigglesDBC.settings.UiFirstLastPage == true) and
      (button == "RightButton")) then
    -- Call the button's 'OnClick' handler until it is disabled
    while (self:IsEnabled()) do
      _G[self:GetName().."_OnClick"](self, button);
    end
  end

  return;
end

--
-- Function to hook into the Garrison UI
--
function NigglesGarrisonUIOnLoad()
  -- Hook functions in the Mission Page of the Garrison Mission panel
  GarrisonMissionFrame.MissionTab.MissionPage:EnableMouse(true);
  GarrisonMissionFrame.MissionTab.MissionPage:HookScript("OnMouseUp",
    function(self, button)
      if ((NigglesDBC.settings.UiGarrisonMission == true) and
          (button == "RightButton")) then
        GarrisonMissionPage_Close();
      end
    end);

  return;
end

--
-- Function to hook into the standard UI
--
-- NOTE: This function should only ever be called once
--
function NigglesHookUi()
  -- Hook functions into the Zoom buttons on the Minimap
  MinimapZoomIn:HookScript("OnMouseUp", minimapZoomIn);
  MinimapZoomOut:HookScript("OnMouseUp", minimapZoomOut);

  -- Hook the clicking of any container item button
  hooksecurefunc("ContainerFrameItemButton_OnModifiedClick",
    bagItemClicked);
  hooksecurefunc("BankFrameItemButtonGeneric_OnModifiedClick",
    bankItemClicked);

  -- Hook functions into the 'Stack Split' frame
  hooksecurefunc("OpenStackSplitFrame", stackSplitOpen);

  -- Hook a function into the equipment buttons
  hooksecurefunc("PaperDollItemSlotButton_OnModifiedClick",
    paperDollItemSlotClicked);

  -- Hook into the Game Time frame
  GameTimeFrame:HookScript("OnClick", NigglesCalendarEventsOnClick);

  -- Hook into the creation of the Achievement Alert frames
  hooksecurefunc("AchievementAlertFrame_OnClick",
    NigglesAchievementsAlertOnClick);

  -- Hook into the selection of a race for Archaeology
  hooksecurefunc("SetSelectedArtifact",
    NigglesArchaeologyOnSelect);

  -- Hook the game tooltip for unit auras and buffs
  hooksecurefunc(GameTooltip, "SetUnitAura", tooltipSetUnitAura);
  hooksecurefunc(GameTooltip, "SetUnitConsolidatedBuff",
    tooltipSetUnitConsolidatedBuff);

  return;
end

--
-- Function to enable/disable features initialised as hooks into standard UI
--
function NigglesHookUiEnable()
  -- Enable/Disable the 'Stack Split Max' frame
  if (StackSplitFrame:IsShown()) then
    stackSplitOpen(stackSplitMaxPurchasable, StackSplitFrame.owner);
  end

  return;
end
